@@ -343,7 +343,7 @@ impl ModuleSpec {
 
     pub fn substitute(
         &self,
-        nodes: &Vec<CircuitRc>,
+        nodes: &[CircuitRc],
         name_prefix: Option<String>,
         name_suffix: Option<String>,
     ) -> Result<CircuitRc> {
@@ -434,7 +434,7 @@ impl ModuleSpec {
 
     pub fn substitute_with_mod_name(
         &self,
-        nodes: &Vec<CircuitRc>,
+        nodes: &[CircuitRc],
         name_prefix: Option<String>,
         name_suffix: Option<String>,
         mod_name: Option<Name>,
@@ -506,7 +506,7 @@ impl ModuleSpec {
                 }
 
                 std::iter::once(recurse(m.spec.circuit.clone(), &SymbolSetRc::new(removed)))
-                    .chain(m.nodes.iter().map(|x| recurse(x.clone(), symbols)))
+                    .chain(m.args().map(|x| recurse(x.clone(), symbols)))
                     .collect::<Result<Vec<_>>>()
             } else {
                 circ.children().map(|x| recurse(x, symbols)).collect()
@@ -727,7 +727,7 @@ impl ModuleSpec {
 
         let circuit = self
             .substitute(
-                &arg_specs.iter().map(|x| x.symbol.crc()).collect(),
+                &arg_specs.iter().map(|x| x.symbol.crc()).collect::<Vec<_>>(),
                 None,
                 None,
             )
@@ -880,13 +880,12 @@ impl ModuleArgSpec {
 /// aka call site
 #[pyclass(extends=PyCircuitBase)]
 #[derive(Clone)]
+#[repr(C)] // for some reason rustc wants to put info last (it's first in all other nodes!), repr(C) stops that
 pub struct Module {
-    #[pyo3(get)]
-    pub nodes: Vec<CircuitRc>,
+    info: CachedCircuitInfo,
+    // invariant: spec.circuit == info.children[0]
     #[pyo3(get)]
     pub spec: ModuleSpec,
-    substitute_cached_info: Box<CachedCircuitInfo>, // box for size
-    info: CachedCircuitInfo,
 }
 
 impl Module {
@@ -922,18 +921,33 @@ impl Module {
         }
         substitute_cached_info.free_symbols = Some(Arc::new(free_syms));
 
-        let substitute_cached_info = Box::new(substitute_cached_info);
-
-        let mut out = Self {
-            nodes,
+        let out = Self {
+            info: CachedCircuitInfo {
+                name,
+                children: once(spec.circuit.clone()).chain(nodes).collect(),
+                ..substitute_cached_info
+            },
             spec,
-            substitute_cached_info,
-            info: Default::default(),
         };
-        out.info.name = name;
         out.initial_init_info()
     }
 
+    pub fn args(&self) -> std::slice::Iter<'_, CircuitRc> {
+        self.info.children[1..].iter()
+    }
+
+    pub fn args_cloned(&self) -> Vec<CircuitRc> {
+        self.info.children[1..].to_vec()
+    }
+
+    pub fn args_slice(&self) -> &[CircuitRc] {
+        &self.info.children[1..]
+    }
+
+    pub fn num_args(&self) -> usize {
+        self.info.children.len() - 1
+    }
+
     pub fn new_kwargs(
         kwargs: &HashMap<Name, CircuitRc>,
         spec: ModuleSpec,
@@ -982,7 +996,7 @@ impl Module {
     fn child_axis_map_inputs_uncached(&self) -> Result<Vec<Vec<Option<usize>>>> {
         let spec_resized = self
             .spec
-            .resize(self.nodes.iter().map(|x| x.info().shape.clone()).collect())?;
+            .resize(self.args().map(|x| x.info().shape.clone()).collect())?;
         let spec_circuit_resized = spec_resized.circuit.clone();
         assert!(spec_circuit_resized.info().shape[..] == self.info().shape[..]);
         let stripped = deep_strip_axis_names(spec_circuit_resized, &None);
@@ -993,11 +1007,8 @@ impl Module {
                 .collect(),
             true,
         );
-        let mut result: Vec<Vec<Option<usize>>> = self
-            .nodes
-            .iter()
-            .map(|c| vec![None; c.info().rank()])
-            .collect();
+        let mut result: Vec<Vec<Option<usize>>> =
+            self.args().map(|c| vec![None; c.info().rank()]).collect();
         visit_circuit_unwrap(out_named, |x| {
             if let Some(sym) = x.as_symbol() {
                 if let Some(i) = self.spec.arg_specs.iter().position(|x| {
@@ -1023,9 +1034,6 @@ impl CircuitNodeHashItems for Module {
 
 impl CircuitNodeSetNonHashInfo for Module {
     fn set_non_hash_info(&mut self) -> Result<()> {
-        let old_name = self.info().name;
-        *self.info_mut() = (*self.substitute_cached_info).clone();
-        self.info_mut().name = old_name;
         Ok(())
     }
 }
@@ -1040,56 +1048,27 @@ impl CircuitNode for Module {
             .collect()
     }
 
-    fn children(&self) -> Box<dyn Iterator<Item = CircuitRc> + '_> {
-        Box::new(std::iter::once(self.spec.circuit.clone()).chain(self.nodes.clone()))
-    }
-
-    fn non_free_children(&self) -> Box<dyn Iterator<Item = CircuitRc> + '_> {
-        Box::new(self.nodes.iter().cloned())
-    }
-
-    fn map_children_enumerate<F>(&self, mut f: F) -> Result<Self>
-    where
-        F: FnMut(usize, CircuitRc) -> Result<CircuitRc>,
-    {
-        let new_spec_circ = f(0, self.spec.circuit.clone())?;
+    fn _replace_children(&self, children: Vec<CircuitRc>) -> Result<Self> {
+        let new_spec = if children[0] == self.spec.circuit {
+            self.spec.clone()
+        } else {
+            ModuleSpec {
+                circuit: children[0].clone(),
+                arg_specs: self.spec.arg_specs.clone(),
+            }
+        };
 
         assert_eq!(
             self.spec.arg_specs.len(),
-            self.nodes.len(),
+            self.num_args(),
             "guaranteed by constructor via expand"
         );
 
-        let nodes = self
-            .nodes
-            .clone()
-            .into_iter()
-            .enumerate()
-            .map(|(i, node)| f(i + 1, node))
-            .collect::<Result<_>>()?;
-
-        Self::try_new(
-            nodes,
-            ModuleSpec {
-                circuit: new_spec_circ,
-                arg_specs: self.spec.arg_specs.clone(),
-            },
-            self.info().name,
-        )
+        Self::try_new(children[1..].to_vec(), new_spec, self.info().name)
     }
 
-    fn map_non_free_children_enumerate<F>(&self, mut f: F) -> Result<Self>
-    where
-        F: FnMut(usize, CircuitRc) -> Result<CircuitRc>,
-    {
-        let nodes = self
-            .nodes
-            .iter()
-            .cloned()
-            .enumerate()
-            .map(|(i, x)| f(i, x))
-            .collect::<Result<_>>()?;
-        Self::try_new(nodes, self.spec.clone(), self.info().name)
+    fn num_free_children(&self) -> usize {
+        1
     }
 
     fn eval_tensors(
@@ -1113,8 +1092,7 @@ impl CircuitNodeAutoName for Module {
                 (self.spec.circuit.info().name.unwrap().string()
                     + "("
                     + &self
-                        .nodes
-                        .iter()
+                        .args()
                         .map(|node| node.info().name.unwrap().into())
                         .collect::<Vec<String>>()
                         .join(", ")
@@ -1169,22 +1147,27 @@ impl Module {
     ) -> Result<Module> {
         let prefix = self.get_prefix(prefix, use_self_name_as_prefix)?;
         Ok(Self::new(
-            self.nodes.clone(),
+            self.args_cloned(),
             self.spec.rename_on_replaced_path(prefix, suffix),
             self.info().name,
         ))
     }
 
+    #[getter(nodes)]
+    pub fn nodes_py(&self) -> Vec<CircuitRc> {
+        self.args_cloned()
+    }
+
     pub fn map_on_replaced_path(&self, f: PyCallable) -> Result<Self> {
         Ok(Self::new(
-            self.nodes.clone(),
+            self.args_cloned(),
             self.spec.py_map_on_replaced_path(f)?,
             self.info().name,
         ))
     }
 
     pub fn aligned_batch_shape(&self) -> Shape {
-        self.spec.aligned_batch_shape(&self.nodes).unwrap()
+        self.spec.aligned_batch_shape(self.args_slice()).unwrap()
     }
 
     /// None is conform all dims
@@ -1197,10 +1180,10 @@ impl Module {
             bail!("dims to conform is larger than num batch dims!")
         }
 
-        let batch_shapes = self.spec.batch_shapes(&self.nodes);
+        let batch_shapes = self.spec.batch_shapes(self.args_slice());
 
         Ok(Self::new(
-            self.nodes.clone(),
+            self.args_cloned(),
             self.spec
                 .resize(
                     self.spec
@@ -1226,9 +1209,9 @@ impl Module {
     // TODO: add some naming options maybe
     pub fn conform_to_input_shapes(&self) -> Self {
         Self::new(
-            self.nodes.clone(),
+            self.args_cloned(),
             self.spec
-                .resize(self.nodes.iter().map(|x| x.shape().clone()).collect())
+                .resize(self.args().map(|x| x.shape().clone()).collect())
                 .expect("constructor should ensure this works"),
             self.info().name,
         )
@@ -1237,7 +1220,7 @@ impl Module {
     fn child_axis_map_inputs(&self) -> Result<Vec<Vec<Option<usize>>>> {
         let key = (
             self.spec.get_hashable(),
-            self.nodes.iter().map(|x| x.info().shape.clone()).collect(),
+            self.args().map(|x| x.info().shape.clone()).collect(),
         );
         if let Some(result) =
             MODULE_EXPANSIONS_AXIS_MAPS.with(|cache| cache.borrow().get(&key).cloned())
@@ -1251,9 +1234,8 @@ impl Module {
     }
 
     pub fn arg_items(&self) -> Vec<(CircuitRc, ModuleArgSpec)> {
-        self.nodes
-            .clone()
-            .into_iter()
+        self.args()
+            .cloned()
             .zip(self.spec.arg_specs.clone())
             .collect()
     }
@@ -1270,7 +1252,7 @@ impl Module {
         name_suffix: Option<String>,
     ) -> CircuitRc {
         self.spec
-            .substitute(&self.nodes, name_prefix, name_suffix)
+            .substitute(self.args_slice(), name_prefix, name_suffix)
             .expect("constructor should ensure this works")
     }
 
@@ -1280,7 +1262,12 @@ impl Module {
         name_suffix: Option<String>,
     ) -> CircuitRc {
         self.spec
-            .substitute_with_mod_name(&self.nodes, name_prefix, name_suffix, self.info().name)
+            .substitute_with_mod_name(
+                self.args_slice(),
+                name_prefix,
+                name_suffix,
+                self.info().name,
+            )
             .expect("constructor should ensure this works")
     }
 
