@@ -28,7 +28,7 @@ use rr_util::{
     pycall, python_error_exception, sv,
     tensor_util::{Shape, TorchDevice, TorchDeviceDtype},
     timed, unwrap,
-    util::{arc_ref_clone, with_context_failable, HashBytes},
+    util::{apply_fn_until_same, arc_ref_clone, with_context_failable, HashBytes},
 };
 use rustc_hash::{FxHashMap as HashMap, FxHashSet as HashSet};
 use thiserror::Error;
