@@ -89,9 +89,8 @@ pub fn unique_to_appearance<T: Eq + Hash + Clone>(vec: &Vec<T>) -> HashMap<T, Ve
     map
 }
 
-pub fn filter_to_idx<T>(col: &Vec<T>, f: &dyn Fn(&T) -> bool) -> Vec<usize> {
-    col.iter()
-        .enumerate()
+pub fn filter_to_idx<T>(col: impl Iterator<Item = T>, f: impl Fn(&T) -> bool) -> Vec<usize> {
+    col.enumerate()
         .filter(|(_i, x)| f(x))
         .map(|(i, _x)| i)
         .collect()
